import io
import base64
from uuid import UUID
from typing import Iterator, Iterable, List, Optional, Union, TYPE_CHECKING
from concurrent.futures import ThreadPoolExecutor
from threading import Event

from PIL import Image as ImagePIL

if TYPE_CHECKING:
    from lqs.common.facade import CoreFacade
from .utils import (
    get_relative_object_path,
    decompress_chunk_bytes,
    get_record_image,
)
from lqs.transcode import Transcode
from lqs.interface.core.models import (
    Ingestion,
    Object,
    Record,
    Topic,
)
from .record_fetcher import RecordFetcher


class RecordUtils:
    def __init__(self, app: "CoreFacade"):
        self.app = app

    def get_record_image(
        self,
        record_data: dict | bytes,
        max_size: Optional[int] = None,
        format: str = "WEBP",
        format_params: dict = {},
        renormalize: bool = False,
        reset_position: bool = True,
        return_bytes: bool = False,
        **kwargs,
    ) -> Union[ImagePIL.Image, io.BytesIO, None]:
        """
        A convenience method which takes deserialized record data from a standard image topic and returns the image as a PIL Image or BytesIO object.

        :param record_data: The record data.
        :type record_data: dict | bytes
        :param max_size: The maximum width or height to downscale to. Defaults to None, which means no downscaling.
        :type max_size: int, optional
        :param format: The output format to use. Defaults to "WEBP".
        :type format: str, optional
        :param format_params: The format parameters to use. Defaults to {}.
        :type format_params: dict, optional
        :param renormalize: Whether to renormalize the image, which is necessary for visualization in some cases. Defaults to False.
        :type renormalize: bool, optional
        :param reset_position: Whether to reset the position offset position of the BytesIO object. Defaults to True.
        :type reset_position: bool, optional
        :param return_bytes: Whether to return the image as a BytesIO object. Defaults to False.
        :type return_bytes: bool, optional
        :returns: The image, either as a PIL Image or BytesIO object, or None if the record data does not contain an image.
        :rtype: Union[ImagePIL.Image, io.BytesIO, None]
        """
        return get_record_image(
            record_data=record_data,
            max_size=max_size,
            format=format,
            format_params=format_params,
            renormalize=renormalize,
            reset_position=reset_position,
            return_bytes=return_bytes,
            **kwargs,
        )

    def iter_topic_records(
        self,
        topic: Union[Topic, UUID, str],
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        per_request_limit: int = 1000,
        frequency: Optional[float] = None,
        include_auxiliary_data: bool = False,
    ):
        """
        Iterate over records for a topic.

        :param topic: The topic to use.
        :type topic: Topic | UUID | str
        :param start_time: The start time to use. Defaults to None.
        :type start_time: int, optional
        :param end_time: The end time to use. Defaults to None.
        :type end_time: int, optional
        :param per_request_limit: The limit to use for each request. Defaults to 1000.
        :type per_request_limit: int, optional
        :param frequency: The frequency to use for each request. Defaults to None.
        :type frequency: float, optional
        :param include_auxiliary_data: Whether to include auxiliary data. Defaults to False.
        :type include_auxiliary_data: bool, optional
        :yields: The record.
        :rtype: Record
        """
        # TODO: there's a bug when using frequency with pagination where records can be skipped
        if isinstance(topic, Topic):
            topic_id = topic.id
        else:
            topic_id = topic
        with ThreadPoolExecutor() as executor:
            records = self.app.list.record(
                topic_id=topic_id,
                timestamp_gte=start_time,
                timestamp_lte=end_time,
                limit=per_request_limit,
                frequency=frequency,
                include_auxiliary_data=include_auxiliary_data,
            ).data
            if len(records) == 0:
                return
            last_record = records[-1]
            last_record_timestamp = last_record.timestamp
            records_future = executor.submit(
                self.app.list.record,
                topic_id=topic_id,
                timestamp_gt=last_record_timestamp,
                timestamp_lte=end_time,
                limit=per_request_limit,
                frequency=frequency,
                include_auxiliary_data=include_auxiliary_data,
            )
            while len(records) > 0:
                yield from records
                records_res = records_future.result()
                records = records_res.data
                if len(records) == 0:
                    break
                last_record = records[-1]
                last_record_timestamp = last_record.timestamp
                records_future = executor.submit(
                    self.app.list.record,
                    topic_id=topic_id,
                    timestamp_gt=last_record_timestamp,
                    timestamp_lte=end_time,
                    limit=per_request_limit,
                    frequency=frequency,
                    include_auxiliary_data=include_auxiliary_data,
                )

    def iter_topics_records(
        self,
        topics: List[Union[Topic, UUID, str]],
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        per_request_limit: int = 1000,
        frequency: Optional[float] = None,
        include_auxiliary_data: bool = False,
    ):
        """
        Iterate over records for multiple topics.

        :param topics: The topics to use.
        :type topics: List[Topic | UUID | str]
        :param start_time: The start time to use. Defaults to None.
        :type start_time: int, optional
        :param end_time: The end time to use. Defaults to None.
        :type end_time: int, optional
        :param per_request_limit: The limit to use for each request. Defaults to 1000.
        :type per_request_limit: int, optional
        :param frequency: The frequency to use for each request. Defaults to None.
        :type frequency: float, optional
        :param include_auxiliary_data: Whether to include auxiliary data. Defaults to False.
        :type include_auxiliary_data: bool, optional
        :yields: The record.
        :rtype: Record
        """
        topic_ids = []
        for topic in topics:
            if isinstance(topic, Topic):
                topic_ids.append(topic.id)
            else:
                topic_ids.append(topic)
        record_iters = {
            topic_id: self.iter_topic_records(
                topic=topic_id,
                start_time=start_time,
                end_time=end_time,
                per_request_limit=per_request_limit,
                frequency=frequency,
                include_auxiliary_data=include_auxiliary_data,
            )
            for topic_id in topic_ids
        }
        next_records = {topic_id: None for topic_id in topic_ids}
        while True:
            for topic_id, record_iter in record_iters.items():
                if next_records[topic_id] is None:
                    next_records[topic_id] = next(record_iter, None)
            next_topic_id = None
            next_record = None
            for topic_id, record in next_records.items():
                if record is not None:
                    if next_record is None or record.timestamp < next_record.timestamp:
                        next_topic_id = topic_id
                        next_record = record
            if next_record is None:
                break
            yield next_record
            next_records[next_topic_id] = next(record_iters[next_topic_id], None)

    def load_auxiliary_data_image(self, source: Union[Record, dict]):
        if isinstance(source, Record):
            auxiliary_data = source.get_auxiliary_data()
        else:
            auxiliary_data = source

        if auxiliary_data is None:
            return None
        if "image" not in auxiliary_data:
            return None
        encoded_webp_data = auxiliary_data["image"]
        decoded_webp_data = base64.b64decode(encoded_webp_data)
        image = ImagePIL.open(io.BytesIO(decoded_webp_data))
        return image

    def get_deserialized_record_data(
        self,
        record: Record,
        topic: Optional[Topic] = None,
        ingestion: Optional[Ingestion] = None,
        transcoder: Optional[Transcode] = None,
    ) -> dict:
        if transcoder is None:
            transcoder = Transcode()

        if topic is None:
            topic = self.app.fetch.topic(record.topic_id).data

        message_bytes = self.fetch_record_bytes(record=record, ingestion=ingestion)

        return transcoder.deserialize(
            type_encoding=topic.type_encoding,
            type_name=topic.type_name,
            type_data=topic.type_data,
            message_bytes=message_bytes,
        )

    def fetch_record_bytes(
        self,
        record: Record,
        ingestion: Optional[Ingestion] = None,
        decompress_chunk: bool = True,
        return_full_chunk: bool = False,
    ) -> bytes:

        if ingestion is None:
            ingestion = self.app.fetch.ingestion(record.ingestion_id).data

        object_store_id = (
            str(ingestion.object_store_id)
            if ingestion.object_store_id is not None
            else None
        )
        object_key = str(ingestion.object_key)

        if record.source is not None:
            # if the record has a source, we need to get the relative path from the object_key
            object_key = get_relative_object_path(
                object_key=object_key, source=record.source
            )

        if object_store_id is None:
            # the data is coming from a log object
            message_bytes: bytes = self.app.fetch.log_object(
                object_key=object_key,
                log_id=record.log_id,
                redirect=True,
                offset=record.data_offset,
                length=record.data_length,
            )
        else:
            # the data is coming from an object store
            message_bytes: bytes = self.app.fetch.object(
                object_key=object_key,
                object_store_id=object_store_id,
                redirect=True,
                offset=record.data_offset,
                length=record.data_length,
            )

        if record.chunk_compression is not None and record.chunk_compression not in [
            "",
            "none",
        ]:
            if decompress_chunk:
                # if the record is compressed, we need to decompress it
                message_bytes = decompress_chunk_bytes(
                    chunk_bytes=message_bytes,
                    chunk_compression=record.chunk_compression,
                    chunk_length=record.chunk_length,
                )
                if not return_full_chunk:
                    # we only return the relevant part of the chunk
                    message_bytes = message_bytes[
                        record.chunk_offset : record.chunk_offset + record.chunk_length
                    ]
            else:
                if not return_full_chunk:
                    raise Exception(
                        "Cannot return partial chunk without decompressing it."
                    )

        return message_bytes

    def get_record_set(
        self,
        records: Iterable[Record],
        carryover_record: Optional[Record] = None,
        density_threshold: float = 0.9,
        max_contiguous_size: int = 100 * 1000 * 1000,  # 100 MB
        max_contiguous_records: int = 1000,
    ) -> tuple[list[list[Record]], Optional[Record]]:
        record_set: list[Record] = []
        relevant_length = 0
        full_length = 0
        start_offset = None
        last_ingestion_id = None
        last_source = None
        last_offset = None

        if carryover_record is not None:
            record_set.append(carryover_record)
            start_offset = carryover_record.data_offset
            last_ingestion_id = carryover_record.ingestion_id
            last_source = carryover_record.source
            last_offset = carryover_record.data_offset + carryover_record.data_length
            carryover_record = None
        leftover_record: Optional[Record] = None

        for record in records:
            if start_offset is None:
                start_offset = record.data_offset
            if last_ingestion_id is None:
                last_ingestion_id = record.ingestion_id
            if last_source is None:
                last_source = record.source
            if last_offset is None:
                last_offset = record.data_offset + record.data_length

            if (record.data_offset + record.data_length) < last_offset:
                # ensure records are ordered properly by offset
                leftover_record = record
                break

            relevant_length += record.data_length
            full_length = record.data_offset + record.data_length - start_offset
            if (
                relevant_length / full_length > density_threshold
                and last_ingestion_id == record.ingestion_id
                and last_source == record.source
                and len(record_set) < max_contiguous_records
                and full_length < max_contiguous_size
                and record.data_offset + record.data_length >= last_offset
            ):
                record_set.append(record)
                last_offset = record.data_offset + record.data_length
            else:
                if len(record_set) == 0:
                    raise Exception("Record set cannot be empty.")
                leftover_record = record
                break
        return record_set, leftover_record

    def get_presigned_url(
        self,
        object_key: str,
        object_store_id: Union[UUID, str, None] = None,
        log_id: Union[UUID, str, None] = None,
        start_offset: Optional[int] = None,
        end_offset: Optional[int] = None,
    ) -> str:
        params = dict(object_key=object_key, redirect=False)
        if object_store_id is None:
            # the data is coming from a log object
            params["log_id"] = log_id
            if start_offset is not None:
                params["offset"] = start_offset
                if end_offset is not None:
                    params["length"] = end_offset - start_offset
            object_meta: Object = self.app.fetch.log_object(**params).data
        else:
            # the data is coming from an object store
            params["object_store_id"] = object_store_id
            if start_offset is not None:
                params["offset"] = start_offset
                if end_offset is not None:
                    params["length"] = end_offset - start_offset
            object_meta: Object = self.app.fetch.object(**params).data
        presigned_url = object_meta.presigned_url
        return presigned_url

    def iter_record_data(
        self,
        records: Iterable[Record],
        deserialize_results: bool = False,
        transcoder: Optional[Transcode] = None,
        density_threshold: float = 0.9,
        max_contiguous_size: int = 100 * 1000 * 1000,
        max_contiguous_records: int = 1000,
        max_workers: Optional[int] = 2,
        order_by_timestamp: bool = True,
        stop_event: Optional[Event] = None,
        use_cache: bool = False,
        cache_dir: str = "/tmp/lqs",
        read_from_cache: bool = True,
        write_to_cache: bool = True,
    ) -> Iterator[tuple[Record, Union[bytes, dict]]]:
        """
        Given a set of records, yield the record and its data.

        :param records: The records to use.
        :type records: Iterable[Record]
        :param deserialize_results: Whether to deserialize the results. Defaults to False.
        :type deserialize_results: bool, optional
        :param transcoder: The transcoder to use. Defaults to None.
        :type transcoder: Transcode, optional
        :param density_threshold: The density threshold to use. Defaults to 0.9.
        :type density_threshold: float, optional
        :param max_contiguous_size: The maximum contiguous size to use. Defaults to 100 * 1000 * 1000.
        :type max_contiguous_size: int, optional
        :param max_contiguous_records: The maximum contiguous records to use. Defaults to 1000.
        :type max_contiguous_records: int, optional
        :param max_workers: The maximum number of workers to use. Defaults to 2.
        :type max_workers: int | None, optional
        :param order_by_timestamp: Whether to order the records by timestamp. Defaults to True.
        :type order_by_timestamp: bool, optional
        :param stop_event: An event to signal stopping the iteration. Defaults to None.
        :type stop_event: Event, optional
        :yields: The record and the record data.
        :rtype: tuple[Record, dict | bytes]
        :param use_cache: Whether to use caching for record data. Defaults to False.
        :type use_cache: bool, optional
        :param cache_dir: The directory to use for caching. Defaults to "/tmp/lqs".
        :type cache_dir: str, optional
        :param read_from_cache: Whether to read from cache if available. Defaults to True.
        :type read_from_cache: bool, optional
        :param write_to_cache: Whether to write to cache after fetching. Defaults to True.
        :type write_to_cache: bool, optional
        """
        fetcher = RecordFetcher(
            app=self.app,
            max_workers=max_workers or 2,
        )
        yield from fetcher.iter_record_data(
            records=records,
            deserialize_results=deserialize_results,
            transcoder=transcoder,
            density_threshold=density_threshold,
            max_contiguous_size=max_contiguous_size,
            max_contiguous_records=max_contiguous_records,
            order_by_timestamp=order_by_timestamp,
            stop_event=stop_event,
            use_cache=use_cache,
            cache_dir=cache_dir,
            read_from_cache=read_from_cache,
            write_to_cache=write_to_cache,
        )



import io
import time
import datetime
from uuid import UUID
from typing import Iterator, Optional, Union, TYPE_CHECKING
from concurrent.futures import ThreadPoolExecutor

import requests

from lqs.transcode import Transcode

if TYPE_CHECKING:
    from lqs.common.facade import CoreFacade
from lqs.interface.core.models import (
    Digestion,
    DigestionPart,
    DigestionPartIndexEntry,
    Record,
    ProcessState,
)


class DigestionUtils:
    def __init__(self, app: "CoreFacade"):
        self.app = app

    def digestion_part_index_entry_to_record(
        self,
        entry: tuple,
        log_id: Union[UUID, str] = "00000000-0000-0000-0000-000000000000",
        created_at: datetime.datetime = datetime.datetime.now(),
    ) -> Record:
        (
            topic_id,
            ingestion_id,
            source,
            data_offset,
            data_length,
            chunk_compression,
            chunk_offset,
            chunk_length,
            timestamp,
        ) = entry
        record = Record(
            log_id=log_id,
            topic_id=topic_id,
            timestamp=timestamp,
            ingestion_id=ingestion_id,
            data_offset=data_offset,
            data_length=data_length,
            chunk_compression=chunk_compression,
            chunk_offset=chunk_offset,
            chunk_length=chunk_length,
            source=source,
            error=None,
            query_data=None,
            auxiliary_data=None,
            raw_data=None,
            context=None,
            note=None,
            locked=False,
            locked_by=None,
            locked_at=None,
            lock_token=None,
            created_at=created_at,
            updated_at=None,
            deleted_at=None,
            created_by=None,
            updated_by=None,
            deleted_by=None,
        )
        return record

    def iter_digestion_part_records(
        self,
        digestion_part: Union[DigestionPart, UUID, str],
        digestion: Union[Digestion, UUID, str, None] = None,
        digestion_part_wait_duration: int = 60 * 30,
        stream_buffer_length: int = 1_000_000 * 32,
        raise_if_ready: bool = True,
        raise_if_queued: bool = False,
    ) -> Iterator[tuple[DigestionPartIndexEntry, bytes]]:
        if isinstance(digestion_part, DigestionPart):
            if digestion_part.index is None:
                # we have to fetch the digestion part to get the index
                digestion_part = self.app.fetch.digestion_part(
                    digestion_id="00000000-0000-0000-0000-000000000000",
                    digestion_part_id=digestion_part.id,
                ).data
        else:
            digestion_part = self.app.fetch.digestion_part(
                digestion_id="00000000-0000-0000-0000-000000000000",
                digestion_part_id=digestion_part,
            ).data

        if digestion is None:
            digestion = self.app.fetch.digestion(digestion_part.digestion_id).data
        else:
            if digestion.id != digestion_part.digestion_id:
                raise Exception(
                    f"Digestion ID {digestion_part.digestion_id} from digestion part does not match provided digestion ID {digestion.id}"
                )

        current_digestion_wait_time = time.time()
        while digestion_part.state != ProcessState.completed:
            if digestion_part.state == ProcessState.failed:
                raise Exception(
                    f"Digestion part {digestion_part.id} is in a failed state."
                )
            if digestion_part.state == ProcessState.queued:
                if raise_if_queued:
                    raise Exception(
                        f"Digestion part {digestion_part.id} is in a queued state."
                    )
            if digestion_part.state == ProcessState.ready:
                if raise_if_ready:
                    raise Exception(
                        f"Digestion part {digestion_part.id} is in a ready state."
                    )
            if time.time() - current_digestion_wait_time > digestion_part_wait_duration:
                raise Exception(
                    f"Digestion part {digestion_part.id} is not in completed state after waiting {digestion_part_wait_duration} seconds."
                )
            time.sleep(10)
            digestion_part = self.app.fetch.digestion_part(
                digestion_id=digestion.id, digestion_part_id=digestion_part.id
            ).data

        record_blob_key = (
            f"digestions/{digestion.id}/digestion_parts/{digestion_part.id}.bin"
        )
        record_blob_object = self.app.fetch.log_object(
            log_id=digestion.log_id,
            object_key=record_blob_key,
            redirect=False,
        ).data
        r = requests.get(record_blob_object.presigned_url, stream=True)
        r.raise_for_status()
        record_blob_data_stream = io.BufferedReader(r.raw, stream_buffer_length)
        for entry_tuple in digestion_part.index:
            (
                topic_id,
                ingestion_id,
                source,
                data_offset,
                data_length,
                chunk_compression,
                chunk_offset,
                chunk_length,
                timestamp,
            ) = entry_tuple
            record_length = data_length if chunk_length is None else chunk_length
            record_bytes = record_blob_data_stream.read(record_length)
            entry = DigestionPartIndexEntry(
                topic_id=topic_id,
                ingestion_id=ingestion_id,
                source=source,
                data_offset=data_offset,
                data_length=data_length,
                chunk_compression=chunk_compression,
                chunk_offset=chunk_offset,
                chunk_length=chunk_length,
                timestamp=timestamp,
            )
            yield entry, record_bytes

    def iter_digestion_records(
        self,
        digestion: Union[Digestion, UUID, str],
        digestion_part_wait_duration: int = 60 * 30,
        stream_buffer_length: int = 1_000_000 * 32,
        max_workers: Optional[int] = 10,
    ) -> Iterator[tuple[UUID, DigestionPartIndexEntry, bytes]]:
        """
        Iterate over records for a digestion.

        A digestion is a collection of digestion parts, which each contain an index of records. This method returns an iterator which yields a tuple with the following:

        - The digestion part ID.
        - The entry from the digestion part index.
        - The record bytes.

        This is useful for iterating over all records in a digestion, which can be more efficient than iterating over records through the API.
        The `max_workers` parameter determines the number of workers to use for fetching the record data in parallel. Set to `None` to use no threading.

        :param digestion: The digestion or digestion ID to use.
        :type digestion: Digestion | UUID | str
        :param digestion_part_wait_duration: The digestion part wait duration to use. Defaults to 60 * 30.
        :type digestion_part_wait_duration: int, optional
        :param stream_buffer_length: The stream buffer length to use. Defaults to 1_000_000 * 32.
        :type stream_buffer_length: int, optional
        :param max_workers: The maximum number of workers to use. Defaults to 10.
        :type max_workers: int, optional

        :yields: The digestion part ID, the entry, and the record bytes.
        """
        if isinstance(digestion, Digestion):
            digestion_id = digestion.id
        else:
            digestion_id = digestion
            digestion = self.app.fetch.digestion(digestion_id).data

        max_digestion_parts = (
            10_000  # TODO: configurable, raise error if too many parts
        )
        limit = 1000
        offset = 0
        digestion_parts_res = self.app.list.digestion_parts(
            digestion_id=digestion_id, limit=limit, offset=offset
        )
        digestion_parts = digestion_parts_res.data
        total_digestion_parts = digestion_parts_res.count
        if total_digestion_parts > max_digestion_parts:
            raise Exception(
                f"Too many digestion parts ({total_digestion_parts}) for extraction (must be less than {max_digestion_parts})"
            )

        while len(digestion_parts) < total_digestion_parts:
            offset += limit
            digestion_parts_res = self.app.list.digestion_parts(
                digestion_id=digestion_id, limit=limit, offset=offset
            )
            digestion_parts.extend(digestion_parts_res.data)
        digestion_parts_count = len(digestion_parts)
        if digestion_parts_count != total_digestion_parts:
            raise Exception(
                f"Expected {digestion_parts_count} digestion parts, but got {total_digestion_parts}"
            )

        if max_workers is None:
            for part_idx, digestion_part in enumerate(digestion_parts):
                self.app.logger.debug(
                    f"Processing digestion part {part_idx + 1} of {digestion_parts_count}"
                )
                for entry, record_bytes in self.iter_digestion_part_records(
                    digestion_part=digestion_part,
                    digestion=digestion,
                    digestion_part_wait_duration=digestion_part_wait_duration,
                    stream_buffer_length=stream_buffer_length,
                ):
                    yield digestion_part.id, entry, record_bytes
        else:

            def get_digestion_part_record_iter(digestion_part):
                items = []
                for entry, record_bytes in self.iter_digestion_part_records(
                    digestion_part=digestion_part,
                    digestion=digestion,
                    digestion_part_wait_duration=digestion_part_wait_duration,
                    stream_buffer_length=stream_buffer_length,
                ):
                    items.append((digestion_part.id, entry, record_bytes))
                return items

            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                futures = []
                for part_idx, digestion_part in enumerate(digestion_parts):
                    future = executor.submit(
                        get_digestion_part_record_iter, digestion_part
                    )
                    futures.append(future)
                for future_idx in range(len(futures)):
                    future = futures[future_idx]
                    for item in future.result():
                        yield item
                    futures[future_idx] = None

    def iter_digestion_data(
        self,
        digestion: Union[Digestion, UUID, str],
        deserialize_results: bool = False,
        transcoder: Optional[Transcode] = None,
        density_threshold: float = 0.9,
        max_contiguous_size: int = 100 * 1000 * 1000,
        max_contiguous_records: int = 1000,
        max_workers: Optional[int] = 2,
    ):
        if isinstance(digestion, Digestion):
            digestion_id = digestion.id
        else:
            digestion_id = digestion
            digestion = self.app.fetch.digestion(digestion_id).data

        log_id = digestion.log_id
        digestion_parts_res = self.app.list.digestion_parts(
            digestion_id=digestion_id, limit=1000
        )
        digestion_parts = digestion_parts_res.data
        digestion_parts_count = len(digestion_parts)
        if digestion_parts_count != digestion_parts_res.count:
            raise Exception(
                f"Expected {digestion_parts_count} digestion parts, but got {digestion_parts_res.count}"
            )

        def iter_digestion_records():
            with ThreadPoolExecutor(max_workers=2) as executor:
                futures = []
                for digestion_part in digestion_parts:
                    future = executor.submit(
                        self.app.fetch.digestion_part,
                        digestion_part_id=digestion_part.id,
                        digestion_id=digestion_id,
                    )
                    futures.append(future)
                for future in futures:
                    digestion_part = future.result().data
                    part_index = digestion_part.index
                    for entry in part_index:
                        record = self.digestion_part_index_entry_to_record(
                            entry=entry, log_id=log_id
                        )
                        yield record

        yield from self.iter_record_data(
            records=iter_digestion_records(),
            deserialize_results=deserialize_results,
            transcoder=transcoder,
            density_threshold=density_threshold,
            max_contiguous_size=max_contiguous_size,
            max_contiguous_records=max_contiguous_records,
            max_workers=max_workers,
        )



import datetime
from uuid import UUID
from typing import Optional, Union, TYPE_CHECKING

from lqs.transcode import Transcode

if TYPE_CHECKING:
    from lqs.common.facade import CoreFacade
from lqs.interface.core.models import (
    Ingestion,
    IngestionPart,
    Record,
)


class IngestionUtils:
    def __init__(self, app: "CoreFacade"):
        self.app = app

    def ingestion_part_index_entry_to_record(
        self,
        entry: tuple,
        ingestion_id: Union[UUID, str],
        source: Optional[str] = None,
        log_id: Union[UUID, str] = "00000000-0000-0000-0000-000000000000",
        created_at: datetime.datetime = datetime.datetime.now(),
    ) -> Record:
        context = None
        if len(entry) == 7:
            (
                topic_id,
                data_offset,
                data_length,
                chunk_compression,
                chunk_offset,
                chunk_length,
                timestamp,
            ) = entry
        elif len(entry) == 8:
            (
                topic_id,
                data_offset,
                data_length,
                chunk_compression,
                chunk_offset,
                chunk_length,
                timestamp,
                context,
            ) = entry
        else:
            raise ValueError(f"Invalid index entry length: {len(entry)}")
        record = Record(
            log_id=log_id,
            topic_id=topic_id,
            timestamp=timestamp,
            ingestion_id=ingestion_id,
            data_offset=data_offset,
            data_length=data_length,
            chunk_compression=chunk_compression,
            chunk_offset=chunk_offset,
            chunk_length=chunk_length,
            source=source,
            error=None,
            query_data=None,
            auxiliary_data=None,
            raw_data=None,
            context=context,
            note=None,
            locked=False,
            locked_by=None,
            locked_at=None,
            lock_token=None,
            created_at=created_at,
            updated_at=None,
            deleted_at=None,
            created_by=None,
            updated_by=None,
            deleted_by=None,
        )
        return record

    def iter_ingestion_part_data(
        self,
        ingestion_part: Union[IngestionPart, UUID, str],
        ingestion: Optional[Ingestion] = None,
        deserialize_results: bool = False,
        transcoder: Optional[Transcode] = None,
        density_threshold: float = 0.9,
        max_contiguous_size: int = 100 * 1000 * 1000,
        max_contiguous_records: int = 1000,
        max_workers: Optional[int] = 2,
    ):
        if isinstance(ingestion_part, IngestionPart):
            ingestion_part_id = ingestion_part.id
            if ingestion_part.index is None:
                ingestion_part = self.app.fetch.ingestion_part(
                    ingestion_part_id=ingestion_part_id,
                    ingestion_id="00000000-0000-0000-0000-000000000000",
                ).data
        else:
            ingestion_part_id = ingestion_part
            ingestion_part = self.app.fetch.ingestion_part(
                ingestion_part_id=ingestion_part_id,
                ingestion_id="00000000-0000-0000-0000-000000000000",
            ).data

        if ingestion is None:
            ingestion = self.app.fetch.ingestion(ingestion_part.ingestion_id).data

        def iter_ingestion_part_records():
            part_index = ingestion_part.index
            for entry in part_index:
                record = self.ingestion_part_index_entry_to_record(
                    entry=entry,
                    ingestion_id=ingestion_part.ingestion_id,
                    source=ingestion_part.source,
                    log_id=ingestion.log_id,
                )
                yield record

        yield from self.iter_record_data(
            records=iter_ingestion_part_records(),
            deserialize_results=deserialize_results,
            transcoder=transcoder,
            density_threshold=density_threshold,
            max_contiguous_size=max_contiguous_size,
            max_contiguous_records=max_contiguous_records,
            max_workers=max_workers,
        )
